//////////////////////////////////////////////////////////
//File description:
//This file is the implementation of the functions defined in 
//readrecord.h.  For more information about these functions,
//please refer to readrecord.h or refer to the description
//of each function below.
//////////////////////////////////////////////////////////
//Includes
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include "readrecord.h"

/*******************************************
read_GSTD_record:  This function reads in a single record from a GSTD
data set.
Para:
  data_file: File pointer to the file that is being read
  f_data   : The float array containing all the float numbers read
  f_num    : Number of float numbers read
  i_data   : The int array containing all the int numbers read
  i_num    : Number of int numbers read

  return value: Indicate whether the operation is successful
Precondition:
  The file should have been opened and the pointer is stored in data_file
Postcondition:
  file pointer will advance to the next record.  Caller can check the
 return value to know whether the operation is successful
 *******************************************/

bool read_GSTD_record(FILE *data_file, float *f_data, int &f_num, int *i_data, int &i_num)
{
	f_num = i_num = 0;

	bool last_line_comment = true;

	while (last_line_comment)
	{
		last_line_comment = false;

		int single_ch;
		do
		{
			single_ch = fgetc(data_file);

			if (single_ch == EOF) return false;

		}while (!isgraph(single_ch));

		switch (single_ch)
		{
			case '/':
				do
				{
					single_ch = fgetc(data_file);

					if (single_ch == EOF) return false;

				}while (single_ch != '\n');
				
				last_line_comment = true;

				break;

			default:  //This is the start of a new line
				
				int op_res = ungetc(single_ch, data_file);

				if (op_res == EOF)
				{
					printf("Unkown error reading file.\n");
					exit(1);
				}

				if (fscanf(data_file, "%d ", &i_data[i_num]) != 1)
					return false; 

				i_num++;

				for (int i = 0; i < 5 ; i++) 
				{
				    if (fscanf(data_file, "%f ", &f_data[f_num]) != 1)
						return false; 
					
					f_num++;
				}
				
				if (fscanf(data_file, "%d\n", &i_data[i_num]) != 1)
					return false; 

				i_num++;
				
				break;
		}
	}

	return true;
}

/*******************************************
read_QueryGen_record:  This function reads in a single record from a GSTD
data set.
Para:
  data_file: File pointer to the file that is being read
  f_data   : The float array containing all the float numbers read
  f_num    : Number of float numbers read
  i_data   : The int array containing all the int numbers read
  i_num    : Number of int numbers read

  return value: Indicate whether the operation is successful
Precondition:
  The file should have been opened and the pointer is stored in data_file
Postcondition:
  file pointer will advance to the next record.  Caller can check the
 return value to know whether the operation is successful
 *******************************************/
bool read_QueryGen_record(FILE *data_file, float *f_data, int &f_num,
			int *i_data, int &i_num)
{
	f_num = i_num = 0;

	bool last_line_comment = true;

	while (last_line_comment)
	{
		last_line_comment = false;

		int single_ch;
		do
		{
			single_ch = fgetc(data_file);

			if (single_ch == EOF) return false;

		}while (!isgraph(single_ch));

		switch (single_ch)
		{
			case '/':
				do
				{
					single_ch = fgetc(data_file);

					if (single_ch == EOF) return false;

				}while (single_ch != '\n');
				
				last_line_comment = true;

				break;

			default:  //This is the start of a new line
				
				int op_res = ungetc(single_ch, data_file);

				if (op_res == EOF)
				{
					printf("Unkown error reading file.\n");
					exit(1);
				}

				for (int i = 0; i < 4; i++) 
				{
				    if (fscanf(data_file, "%f ", &f_data[f_num]) != 1)
						return false; 
					
					f_num++;
				}
				
				if (fscanf(data_file, "%d ", &i_data[i_num]) != 1)
					return false; 

				i_num++;

				if (fscanf(data_file, "%d\n", &i_data[i_num]) != 1)
					return false; 

				i_num++;
				
				break;
		}
	}

	return true;
}



